//  MUWidget.cpp version 1.5
//  yudit package - Unicode Editor for the X Window System (and Linux) 
//
//  Author: gsinai@iname.com (Gaspar Sinai)
//  GNU Copyright (C) 1997,1998,1999  Gaspar Sinai
// 
//  yudit version 1.5  Copyright(C) 30 November, 1999, Tokyo Japan  Gaspar Sinai
//  yudit version 1.4  Copyright(C) 25 November, 1999, Tokyo Japan  Gaspar Sinai
//  yudit version 1.3  Copyright(C)  5 April,    1999, Tokyo Japan  Gaspar Sinai
//  yudit version 1.2  Copyright(C) 10 December, 1998, Tokyo Japan  Gaspar Sinai
//  yudit version 1.1  Copyright(C) 23 August,   1998, Tokyo Japan  Gaspar Sinai
//  yudit version 1.0  Copyright(C) 17 May,      1998, Tokyo Japan  Gaspar Sinai
//  yudit version 0.99 Copyright(C)  4 April,    1998, Tokyo Japan  Gaspar Sinai
//  yudit version 0.97 Copyright(C)  4 February, 1998, Tokyo Japan  Gaspar Sinai
//  yudit version 0.95 Copyright(C) 10 January,  1998, Tokyo Japan  Gaspar Sinai
//  yudit version 0.94 Copyright(C) 17 December, 1997, Tokyo Japan  Gaspar Sinai
//  yudit version 0.9 Copyright (C)  8 December, 1997, Tokyo Japan  Gaspar Sinai
//  yutex version 0.8 Copyright (C)  5 November, 1997, Tokyo Japan  Gaspar Sinai
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//
//

#include "MUWidget.h"
#include <X11/Xatom.h>
#include <Xm/ScrollBar.h>
#include <Xm/Form.h>
#include <Xm/Frame.h>
#include <Xm/DrawingA.h>
#include <Xm/CutPaste.h>
#include <AString.h>
#include <UList.h>
#include <strings.h>
#include <X11/Intrinsic.h>

static void MUActionEnter (Widget w, XEvent e, String *s, Cardinal *c);
static void MUActionEscape (Widget w, XEvent e, String *s, Cardinal *c);
static void MUActionTab (Widget w, XEvent e, String *s, Cardinal *c);
static void MUActionSTab (Widget w, XEvent e, String *s, Cardinal *c);

static char *commonTranslations = "#override\n\
        <Key>Return: enter()\n\
        <Key>KP_Enter: enter()\n\
        <Key>Tab: tab()\n\
        Shift Meta Alt Ctrl <Key>Tab: stab()\n\
        <Key>Escape: escape() \n\
        <Key>osfEscape: escape() \n\
        <Key>osfCancel: escape() ";

static  XtActionsRec    commonActions[] =
{
        { "enter", (XtActionProc) MUActionEnter },
        { "escape", (XtActionProc) MUActionEscape },
        { "tab", (XtActionProc) MUActionTab },
        { "stab", (XtActionProc) MUActionSTab },

};

#define BLINK_TIME	500 // ms
#define BUFFER_SIZE	1024

#define EVENT_MASK (KeyPressMask | ButtonPressMask | ButtonReleaseMask |\
		ExposureMask | Button1MotionMask | VisibilityChangeMask |\
		FocusChangeMask | StructureNotifyMask ) 

static Boolean MUEditConvertSelection (Widget w, Atom* selection, Atom* target, 
	Atom* type, XtPointer* value, unsigned long *length, int* format);

static UList<Widget> widgetList;

MUWidget::MUWidget (const char* nameIn, Widget parentIn, 
	const UWidget::UEditable editableIn)
	: UWidget (nameIn, MULTILINE_SCROLL,  editableIn), 
	background (parentIn),
	foreground (parentIn),
	cursorColor (parentIn)
{
	parent=parentIn;
	editArea=0;
	editFrame=0;
	hbar=0;
	vbar=0;
	timerOn=0;
	pasteStarted=0;
	clipBoard=0;
	clipBoardLen=0;
	xInputFont=0;
	xInput=0;
	xInputFontMap=0;

	
	hbar = XtVaCreateManagedWidget ("HorizontalScroll", 
		xmScrollBarWidgetClass, parent,
		XmNorientation,		XmHORIZONTAL,
		XmNrightAttachment,	XmATTACH_FORM,
		XmNleftAttachment,	XmATTACH_FORM,
		XmNbottomAttachment,	XmATTACH_FORM,
		XmNvalue,		0,
		XmNincrement,		1,
		XmNsliderSize,		1,
		XmNpageIncrement,	1,
		XmNmaximum,		1,
		0);

	vbar = XtVaCreateManagedWidget ("VerticalScroll", 
		xmScrollBarWidgetClass, parent,
		XmNbottomWidget,	hbar,
		XmNorientation,		XmVERTICAL,
		XmNrightAttachment,	XmATTACH_FORM,
		XmNtopAttachment,	XmATTACH_FORM,
		XmNbottomAttachment,	XmATTACH_WIDGET,
		XmNvalue,		0,
		XmNincrement,		1,
		XmNsliderSize,		1,
		XmNpageIncrement,	1,
		XmNmaximum,		1,
		0);

	// Form can not have input focus - otherwise I would use it .
	editFrame = XtVaCreateManagedWidget ("EditFrame",
		xmFrameWidgetClass,	parent,
		XmNshadowType,		XmSHADOW_IN,
		XmNshadowThickness,	2,
		XmNrightWidget,		vbar,
		XmNbottomWidget,	hbar,
		XmNleftAttachment,	XmATTACH_FORM,
		XmNrightAttachment,	XmATTACH_WIDGET,
		XmNbottomAttachment,	XmATTACH_WIDGET,
		XmNtopAttachment,	XmATTACH_FORM,
		0);

	editArea = XtVaCreateManagedWidget ("EditArea",
		xmDrawingAreaWidgetClass,	editFrame,
		XmNorientation, 	XmVERTICAL,
		0);

	// To prevent motif thingy from doing funny stuff.
	XtAppAddActions (XtWidgetToApplicationContext (editArea),
		commonActions, XtNumber (commonActions));

	XtOverrideTranslations (editArea,
		XtParseTranslationTable (commonTranslations));

	widgetList.addItem (editArea, (char*) this);

}

void 
MUWidget::setupWidget (UWidget *uwidgetIn)
{
	Dimension	width=0;

	XtVaGetValues (vbar, XmNwidth, &width, 0);
	XtVaSetValues (hbar, XmNrightOffset, width, 0);

	// Inherit font, input, tabsize ....
	uSetWidget (XtDisplay (editArea), XtWindow (editArea));
	uInherit (uwidgetIn);

	XtVaSetValues (vbar, XmNminimum, 0,  0);
	XtVaSetValues (hbar, XmNminimum, 0,  0);

	XtAddCallback (vbar, XmNvalueChangedCallback, 
		(XtCallbackProc) MuedVScolled, this);
	XtAddCallback (hbar, XmNvalueChangedCallback, 
		(XtCallbackProc) MuedHScolled, this);
	XtAddCallback (vbar, XmNdragCallback, 
		(XtCallbackProc) MuedVScolled, this);
	XtAddCallback (hbar, XmNdragCallback, 
		(XtCallbackProc) MuedHScolled, this);

	//XtAddCallback (editArea, XmNresizeCallback, (XtCallbackProc) MuedResized, this);

	// They do not work as they should.
	// XtAddCallback (editArea, XmNexposeCallback, (XtCallbackProc) MuedEvent, this);
	// XtAddCallback (editArea, XmNinputCallback, (XtCallbackProc) MuedEvent, this);
	XtAddEventHandler (editArea, 
		EVENT_MASK,
		True,
		(XtEventHandler) MUEditEventHandler, this);
		
	XmProcessTraversal (editArea, XmTRAVERSE_CURRENT);

	setBackground ("white");
	setForeground ("black");
	setCursorColor ("black");
	XFlush (XtDisplay (editArea));

	uRefresh ();
}

// call this after we mapped other widgets.
void
MUWidget::setupWidget (
	const char* fontIn, int fontSizeIn, const char* fontWeightIn,
	const char* fontSlantIn, const char* fontSpacingIn,
	const char* addStyleIn, const char* avgWidthIn)
{
	Dimension	width=0;

	XtVaGetValues (vbar, XmNwidth, &width, 0);
	XtVaSetValues (hbar, XmNrightOffset, width, 0);

	uSetWidget (XtDisplay (editArea), XtWindow (editArea));

	XtVaSetValues (vbar, XmNminimum, 0,  0);
	XtVaSetValues (hbar, XmNminimum, 0,  0);

	XtAddCallback (vbar, XmNvalueChangedCallback, 
		(XtCallbackProc) MuedVScolled, this);
	XtAddCallback (hbar, XmNvalueChangedCallback, 
		(XtCallbackProc) MuedHScolled, this);
	XtAddCallback (vbar, XmNdragCallback, 
		(XtCallbackProc) MuedVScolled, this);
	XtAddCallback (hbar, XmNdragCallback, 
		(XtCallbackProc) MuedHScolled, this);

	//XtAddCallback (editArea, XmNresizeCallback, MuedResized, this);

	// They do not work as they should.
	// XtAddCallback (editArea, XmNexposeCallback, MuedEvent, this);
	// XtAddCallback (editArea, XmNinputCallback, MuedEvent, this);
	XtAddEventHandler (editArea, 
		EVENT_MASK,
		True,
		(XtEventHandler) MUEditEventHandler, this);
		
	XmProcessTraversal (editArea, XmTRAVERSE_CURRENT);

	setBackground ("white");
	setForeground ("black");
	setCursorColor ("black");
	XFlush (XtDisplay (editArea));
	uSetFont (fontIn);
	uSetFontSize (fontSizeIn);
	uSetFontWeight (fontWeightIn);
	uSetFontSlant (fontSlantIn);
	uSetFontSpacing (fontSpacingIn);
	uSetFontAddStyle (addStyleIn);
	uSetFontAvgWidth (avgWidthIn);
	uRefresh ();
}

MUWidget::~MUWidget ()
{
	int 		looped;
	XtAppContext	cxt;
	XEvent		event;
	UWidget*	uwidget;

	if (xInputFont!=0)
	{
		delete xInputFont;
		xInputFont = 0;
	}
	if (xInputFontMap!=0)
	{
		// We may use unusefontamp later.
		xInputFontMap = 0;
	}
	uwidget = this;
	if (xInput!=0)
	{
		xInput->uEnd();
		uwidget->uSetXInput ((UXInput*) 0);
		switch (xInputType)
		{
		case KinputType:
			delete ((MUXKinput*) xInput);
			break;
		default:
			delete xInput;
			break;
		}
		xInput=0;
	}

	cxt = XtWidgetToApplicationContext (editArea);
	if (timerOn)
	{
		XtRemoveTimeOut (timerId);
	}
	looped = 0;
	if (pasteStarted) cerr << "info: waiting for paste to finish...\n";
	while (pasteStarted)
	{
		XtAppNextEvent(cxt, &event);
		XtDispatchEvent(&event);
		looped++;
	}
	if (looped!=0) cerr << "info: paste finished after " << looped 
		<< " events.";

	XtRemoveEventHandler (editArea, 
		EVENT_MASK,
		True,
		(XtEventHandler) MUEditEventHandler, this);

	widgetList.deleteItem (editArea);
	if (editArea!=0) XtDestroyWidget (editArea);
	if (hbar!=0) XtDestroyWidget(hbar);
	if (vbar!=0) XtDestroyWidget(vbar);
	if (clipBoard!=0)
	{
		delete clipBoard;
		clipBoard=0;
	}
}

#ifdef USE_MOTIF_CLIPBOARD
void
MUWidget::uClipboardPut (const unsigned char * charString, const int len)
{
	XmString	label=0;
	long		itemId;
	long		dataId;
	int		status;
	char*		copyStr;

	copyStr = new char[len + 1 ];
	CHECKNULL (copyStr);
	if (len>0) memcpy (copyStr, charString, len);
	copyStr[len] = 0;

	label = XmStringCreateLtoR ("yudit", XmSTRING_DEFAULT_CHARSET);
	do
	{
		status = XmClipboardStartCopy (
			XtDisplay (parent),
			XtWindow (parent),
			label, CurrentTime,
			editArea, 0, &itemId);
	} while (status == ClipboardLocked);
	XmStringFree (label);
	do
	{
		status = XmClipboardCopy (XtDisplay (parent), 
			XtWindow (parent),
			itemId, "STRING",  copyStr,
			(long) len, 0L, &dataId);
	} while (status == ClipboardLocked);
	delete copyStr;
	
	do
	{
		status = XmClipboardEndCopy (XtDisplay (parent),
			XtWindow (parent),
			itemId);
	} while (status == ClipboardLocked);
		
}

unsigned char*	
MUWidget::uClipboardGet (int* len)
{
	char		buffer[BUFFER_SIZE+1];
	unsigned char*		str;
	unsigned char*		copystr;
	int		strSize;

	unsigned long 	got;
	long		privateId;
	int		status;

	do 
	{
		status = XmClipboardStartRetrieve (
			XtDisplay (parent),
			XtWindow (parent),
			CurrentTime);
	} while (status == ClipboardLocked);

	strSize =0;
	do
	{
		got = 0;
		do 
		{
			status = XmClipboardRetrieve (
				XtDisplay (parent),
				XtWindow (parent),
				"STRING", buffer, BUFFER_SIZE, &got, &privateId);
		} while (status == ClipboardLocked);
		if (got==0) continue;
		copystr = new unsigned char[strSize + got +1];
		CHECKNULL (copystr);
		if (strSize!=0)
		{
			memcpy (copystr, str, strSize);
			delete str;
		}
		memcpy (&copystr[strSize], buffer, got);
		str = copyStr;
		strSize += got;
		
	} while (status == ClipboardTruncate);

	do
	{
		status = XmClipboardEndRetrieve (XtDisplay (parent),
			XtWindow (parent));
	} while (status == ClipboardLocked);

	*len = strSize;
	if (strSize==0) return 0;
	str[strSize]=0;
	return str;
}
#else /* !USE_MOTIF_CLIPBOARD */

void
MUWidget::uClipboardPut (const unsigned char * charString, const int len)
{
	Boolean status;

	// Get ownership
	clipBoard = new unsigned char[len+1];
	if (clipBoard!=0)
	{
		memcpy (clipBoard, charString, len);
		clipBoard[len]=0;
		clipBoardLen = len;
	}
	
	status = XtOwnSelection (editArea, XA_PRIMARY, CurrentTime,
			MUEditConvertSelection, MUEditLoseSelection, 0);
}

unsigned char*	
MUWidget::uClipboardGet (int *len)
{
	int		looped;
	XtAppContext	cxt;
	XEvent		event;
	Atom		compoundText;

	cxt = XtWidgetToApplicationContext (editArea);

	if (pasteStarted) cerr << "info: waiting for paste to finish...\n";
	looped = 0;
	while (pasteStarted)
	{
		XtAppNextEvent(cxt, &event);
		XtDispatchEvent(&event);
		looped++;
	}
	if (looped!=0) cerr << "info: paste finished after " << looped 
		<< " events.";
	pasteStarted = 1;
	compoundText = XInternAtom (XtDisplay (editArea), 
		"COMPOUND_TEXT", False);
	pasteString = 0;
	pasteStringLen=0;

	// Try compoundText
	if (compoundText!=None)
	{
		XtGetSelectionValue (editArea, XA_PRIMARY, compoundText,
			(XtSelectionCallbackProc) MUEditGetSelection, 
			this, CurrentTime);
		while (pasteStarted)
		{
			XtAppNextEvent(cxt, &event);
			XtDispatchEvent(&event);
		}
	}
	if (pasteString==0)
	{
		pasteStarted = 1;
		XtGetSelectionValue (editArea, XA_PRIMARY, XA_STRING,
			(XtSelectionCallbackProc) MUEditGetSelection, 
			this, CurrentTime);
		while (pasteStarted)
		{
			XtAppNextEvent(cxt, &event);
			XtDispatchEvent(&event);
		}
	}
	*len = pasteStringLen;
	return (unsigned char*) pasteString;
	
}

#endif /* !USE_MOTIF_CLIPBOARD */

void
MUWidget::uSetVScroll (int value, int step, int page, int max)
{
	ScrollValue	newValues;

	newValues = vValues;

	if (max !=0 || page != 0)
	{
		newValues.page = (page<=0) ? 1 : page;
		newValues.max = (max<=0) ? 1 : max; 
		newValues.step = (step<=0) ? 1 : step;
		if (newValues.page > newValues.max) newValues.max = newValues.page;
		if (newValues.value>newValues.max-newValues.page)
		{
			newValues.value = newValues.max-newValues.page;
		}
	}
	
	if (value >= 0)
	{
		newValues.value = (value>newValues.max-newValues.page) 
			? newValues.max-newValues.page: value;
	}

	if (newValues.max > vValues.max)
	{ 
		XtVaSetValues (vbar, XmNmaximum, newValues.max, 0);
	}
	XmScrollBarSetValues (vbar, newValues.value,
		newValues.page, newValues.step, newValues.page, False);
	if (newValues.max < vValues.max)
	{ 
		XtVaSetValues (vbar, XmNmaximum, newValues.max, 0);
	}
	if (newValues.value != vValues.value) uVScroll (newValues.value);
	vValues = newValues;
}

void
MUWidget::uSetHScroll (int value, int step, int page, int max)
{
	ScrollValue	newValues;

	newValues = hValues;

	if (max !=0 || page != 0)
	{
		newValues.page = (page<=0) ? 1 : page;
		newValues.max = (max<=0) ? 1 : max; 
		newValues.step = (step<=0) ? 1 : step;
		if (newValues.page > newValues.max) newValues.max = newValues.page;
		if (newValues.value>newValues.max-newValues.page)
		{
			newValues.value = newValues.max-newValues.page;
		}
	}
	
	if (value >= 0)
	{
		newValues.value = (value>newValues.max-newValues.page) 
			? newValues.max-newValues.page: value;
	}

	if (newValues.max > hValues.max)
	{ 
		XtVaSetValues (hbar, XmNmaximum, newValues.max, 0);
	}
	XmScrollBarSetValues (hbar, newValues.value,
		newValues.page, newValues.step, newValues.page, False);
	if (newValues.max < hValues.max)
	{ 
		XtVaSetValues (hbar, XmNmaximum, newValues.max, 0);
	}
	if (newValues.value != hValues.value) uHScroll (newValues.value);
	hValues = newValues;
}

ScrollValue::ScrollValue (void)
{
	value = 0;
	max = 1;
	step = 1;
	page = 1;
}

ScrollValue::~ScrollValue ()
{
}
void 
MuedHScolled (Widget sb, MUWidget *muwidget, XmScrollBarCallbackStruct *cbs)
{
	muwidget->hValues.value = cbs->value;
	muwidget->uHScroll (cbs->value);
}

void 
MuedVScolled (Widget sb, MUWidget *muwidget, XmScrollBarCallbackStruct *cbs)
{
	muwidget->vValues.value = cbs->value;
	muwidget->uVScroll (cbs->value);
}

void 
MuedResized (Widget w, MUWidget *muwidget, XmAnyCallbackStruct *cbs)
{
	Dimension	width=0;
	Dimension	height=0;

	XtVaGetValues (w,
		XmNheight,	&height,
		XmNwidth,	&width,
		0);
	muwidget->uResize ((int)width, (int)height);
}

//
// Absolete
//
void 
MuedEvent (Widget w, MUWidget *muwidget, XmDrawingAreaCallbackStruct *cbs)
{
	muwidget->uHandleEvent (cbs->event);
}

void 
MUEditEventHandler (Widget w, MUWidget* muwidget, XEvent* event, 
	Boolean* callAgain)
{
	if (event->type == FocusIn)
	{
		MuedFocusIn (w, muwidget, 0);
	}
	if (event->type == FocusOut)
	{
		MuedFocusOut (w, muwidget, 0);
	}
	muwidget->uHandleEvent (event);
	*callAgain = True;;
}

void 
MuedFocusIn (Widget w, MUWidget *muwidget, XmAnyCallbackStruct *cbs)
{
	if (muwidget->timerOn)
	{
		return;
	}
	muwidget->timerId = XtAppAddTimeOut (
		XtDisplayToApplicationContext (XtDisplay (muwidget->parent)),
		BLINK_TIME, (XtTimerCallbackProc) MuedTimerCB, muwidget);
	muwidget->timerOn = 1;
}

void 
MuedFocusOut (Widget w, MUWidget *muwidget, XmAnyCallbackStruct *cbs)
{
	if (muwidget->timerOn)
	{
		XtRemoveTimeOut (muwidget->timerId);
		muwidget->timerOn=0;
	}
	muwidget->uBlinkCursor (1);
	return;
}

void
MuedTimerCB (MUWidget *muwidget, XtIntervalId *id)
{
	muwidget->uBlinkCursor ();
	muwidget->timerId = XtAppAddTimeOut (
		XtDisplayToApplicationContext (XtDisplay (muwidget->parent)),
		BLINK_TIME, (XtTimerCallbackProc) MuedTimerCB, muwidget);
}

UString::UStatus
MUWidget::setConversion (const char *conversion)
{
	return uSetConverter (conversion);
}

UWidget::UStatus
MUWidget::setInput (const char *input)
{
	return uSetInput (input);
}

void
MUWidget::setTabSize (int newValue)
{
	uSetTabSize(newValue);
	uRefresh();
}

void
MUWidget::setBackground (const char* color)
{
	background.setNamedColor (color);
	uSetBackground((unsigned long)background.pixel());
	uRefresh();
}

void
MUWidget::setForeground (const char* color)
{
	foreground.setNamedColor (color);
	uSetForeground((unsigned long)foreground.pixel());
	uRefresh();
}

void
MUWidget::setCursorColor (const char* color)
{
	cursorColor.setNamedColor (color);
	uSetCursorColor((unsigned long)cursorColor.pixel());
	uRefresh();
}


void
MUWidget::appendText (const unsigned char *text, const int len)
{
	uAppendText (text, len, 0, 0);
}

unsigned char*
MUWidget::getLine (const int pos, int *len)
{
	
	return uGetLine (pos, len);
}

unsigned char*
MUWidget::getBuffer (int* len)
{
	
	return uGetBuffer (len);
}

void 
MUWidget::refresh()
{
	uRefresh();
}


void
MUEditGetSelection (Widget w, MUWidget* muwidget, Atom* selection, Atom* type, 
	XtPointer value, unsigned long *length, int* format)
{
	if (*type == 0 /*XT_CONVERT_FAIL*/ || *length == 0 || value == NULL)
	{
		cerr << "info: nothing on clipboard.\n";
		muwidget->pasteString = 0;
		muwidget->pasteStarted = 0;
		return;
	}
	if (*format != 8)
	{
		cerr << "info: size is " << *format << ".\n";
		muwidget->pasteString = 0;
		muwidget->pasteStarted = 0;
		XtFree ((char*)value);
		return;
	}
	muwidget->pasteStringLen = *length;
	muwidget->pasteString = new char[*length +1];
	CHECKNULL (muwidget->pasteString);
	if (*length!=0) memcpy (muwidget->pasteString, value, *length);
	muwidget->pasteString[*length] = 0;
	muwidget->pasteStarted = 0;
	XtFree ((char*)value);
	return;
}

void
MUEditLoseSelection (Widget w, Atom* selection)
{
	MUWidget*  muwidget;

	muwidget=(MUWidget*) widgetList.getClient (w);
	if (muwidget==0)
	{
		cerr << "error: convertSelection client data missing from ";
		cerr << (const char*) XtName (w) << ".\n";
		return;
	}
	if (muwidget->clipBoard!=0)
	{
		delete muwidget->clipBoard;
		muwidget->clipBoard=0;
		muwidget->clipBoardLen=0;
	}
	muwidget->uClipboardClear();
}

static Boolean
MUEditConvertSelection (Widget w, Atom* selection, Atom* target, Atom* type, 
	XtPointer* value, unsigned long *length, int* format)
{
	MUWidget*  	muwidget;
	Atom		compoundText;
	Atom		utfText;
	Atom		textType;

	muwidget=(MUWidget*) widgetList.getClient (w);
	if (muwidget==0)
	{
		cerr << "error: convertSelection client data missing from ";
		cerr << (const char*) XtName (w) << ".\n";
		return False;
	}
	compoundText = XInternAtom (XtDisplay (w), "COMPOUND_TEXT", False);
	utfText = XInternAtom (XtDisplay (w), "UTF8_TEXT", False);
	textType = XInternAtom (XtDisplay (w), "TEXT", False);
	
	switch (*target)
	{
	case XA_STRING:
		*type = XA_STRING;
		if (muwidget->clipBoard==0) return False;
		*format=8;
		*length = muwidget->clipBoardLen;;
		*value = XtMalloc (length[0] + 1);
		memcpy (*value, (const char*) muwidget->clipBoard, *length);
		((char*)*value)[*length] = 0;
		return True;
	
	default:
		if (*target != textType && *target !=compoundText && *target !=utfText) 
		{
			XBell (XtDisplay (w), 50);
			return False;
		}
		*type = *target;
		if (muwidget->clipBoard==0) return False;
		*format=8;
		*length = muwidget->clipBoardLen;
		*value = XtMalloc (length[0] + 1);
		memcpy (*value, (const char*) muwidget->clipBoard, *length);
		((char*)*value)[*length] = 0;
		return True;
	}
	return False;
}

int
MUWidget::uSetXInput (const char* _name, const char* imIn, 
	const char* inputStyleIn, const char* textType, 
	const char* uStringIn, const char* fontMapIn, double versionIn)
{
	UXInput::UInputStyle	inputStyle;
	UWidget*		uwidget;

	if (xInputFont!=0)
	{
		delete xInputFont;
		xInputFont = 0;
	}
	if (xInputFontMap!=0)
	{
		// We may use unusefontamp later.
		xInputFontMap = 0;
	}
	uwidget = this;
	if (xInput!=0)
	{
		xInput->uEnd();
		uwidget->uSetXInput ((UXInput*) 0);
		switch (xInputType)
		{
		case KinputType:
			delete ((MUXKinput*) xInput);
			break;
		default:
			delete xInput;
			break;
		}
		xInput=0;
	}
	if (_name==0 || strcasecmp (_name, "None") ==0)
	{
		return 0;
	}
	xInputFontMap = UGetFontMap (fontMapIn);
	if (xInputFontMap==0)
	{
		cerr << "error: can not get fontmap '" << fontMapIn << "'.\n";
		return -1;
	}
	xInputFontMapName = fontMapIn;

	uSetCursor (UCursor::WAIT, 1);
	xInputFont = new UFontX11 (xInputFontMap,
		XtDisplay (editArea), DefaultScreen (XtDisplay(editArea)));
	inputStyle = UXInput::UOFF;
	if (strcasecmp (inputStyleIn, "Root")==0)
	{
		inputStyle = UXInput::UROOT;
	}
	if (strcasecmp (_name, "Kinput2")==0)
	{
		xInputType = KinputType;
		xInput = new MUXKinput (getWidget(), uwidget);
		if (xInput->uSetup (imIn, inputStyle,
			textType, uStringIn, xInputFont, 
			(float) versionIn) != UXInput::OK)
		{
			delete ((MUXKinput*) xInput);
			xInput = 0;
			cerr << "error: can not set up xInput..";
			uSetCursor (UCursor::INSERT, 1);
			return -1;
		}
		MAddXInputCheck (xInput, getWidget());
		uwidget->uSetXInput ((UXInput*) xInput);
		uSetCursor (UCursor::INSERT, 1);
		return 0;
	}
	uSetCursor (UCursor::INSERT, 1);
	return -1;
}

void
MUWidget::setMultiLine (UWidget::UMultiLine multiLineIn)
{
	UWidget*		uwidget;

	uwidget = this;

	uwidget->uSetMultiLine (multiLineIn);
	switch (multiLineIn)
	{
	case UWidget::MULTILINE_SCROLL:
		if (!XtIsManaged (hbar))
		{
			XtManageChild (hbar);
			XtVaSetValues (editFrame,
				XmNbottomWidget, hbar,
				XmNbottomAttachment, XmATTACH_WIDGET,
				0);
		}
		if (!XtIsManaged (vbar))
		{
			XtManageChild (vbar);
			XtVaSetValues (editFrame,
				XmNbottomWidget, vbar,
				XmNrightAttachment, XmATTACH_WIDGET,
				0);
		}
		break;
	case UWidget::SINGLE_SCROLL:
		if (!XtIsManaged (hbar))
		{
			XtManageChild (hbar);
			XtVaSetValues (editFrame,
				XmNbottomWidget, hbar,
				XmNbottomAttachment, XmATTACH_WIDGET,
				0);
		}
		if (XtIsManaged (vbar))
		{
			XtUnmanageChild (vbar);
			XtVaSetValues (editFrame,
				XmNrightAttachment, XmATTACH_FORM,
				0);
		}
		break;
	default:
		if (XtIsManaged (hbar))
		{
			XtUnmanageChild (hbar);
			XtVaSetValues (editFrame,
				XmNbottomAttachment, XmATTACH_FORM,
				0);
		}
		if (XtIsManaged (vbar))
		{
			XtUnmanageChild (vbar);
			XtVaSetValues (editFrame,
				XmNrightAttachment, XmATTACH_FORM,
				0);
		}
	}
}

static void
MUActionEnter (Widget w, XEvent e, String *s, Cardinal *c)
{
	XmProcessTraversal (w,  XmTRAVERSE_NEXT_TAB_GROUP);
}

static void
MUActionEscape (Widget w, XEvent e, String *s, Cardinal *c)
{
}

static void
MUActionTab (Widget w, XEvent e, String *s, Cardinal *c)
{
}

static void
MUActionSTab (Widget w, XEvent e, String *s, Cardinal *c)
{
}
